{{
@file pjvm_demo.spin:

  This is a PASM version of Peter's Spin JVM.
  This code is compatible with much of the Javelin Stamp
  library at http://tech.groups.yahoo.com/group/JavelinCode
  (testing is still in progress, so there may be errors.)

  '------------------------------------------------
  The JVM provides 18KB to 19KB for user Java bytecode.
  This is different from the Javelin which has 32KB.
  '------------------------------------------------

  '------------------------------------------------
  In the future, being able to define custom JNI devices
  is very desirable. LCD, touch screen, micro-ball mouse,
  keyboard/keypad, tone generator, etc... are possible.
  '------------------------------------------------

  '------------------------------------------------
  Minimum COG configuration:
  2 COGs used; 6 COGs free for user apps.

  1 for Spin and VM service handler.
  1 for JVM Core Interpreter.
  '------------------------------------------------

  '------------------------------------------------
  Standard Javelin-like configuration:
  5 COGs used; 3 free for user apps.

  1 for Spin and VM service handler.
  1 for JVM Core Interpreter.
  3 for 6 VPs by default with up to 4 for 8 VPs possible.
  '------------------------------------------------

  '------------------------------------------------
  Other Notes:
  '------------------------------------------------
  Use FDX define to enable high speed serial - uses 1 COG
  Garbage collection is not functional now and disabled.
  '------------------------------------------------

  '------------------------------------------------
  Performance:
  '------------------------------------------------
  A recursive FIBO test gives the following with 80MHz clock

  FIBO(00) =     0 (    6 ticks) (    3 ms)
  FIBO(01) =     1 (    6 ticks) (    3 ms)
  FIBO(02) =     1 (    6 ticks) (    3 ms)
  FIBO(03) =     2 (    6 ticks) (    3 ms)
  FIBO(04) =     3 (    6 ticks) (    3 ms)
  FIBO(05) =     5 (    7 ticks) (    4 ms)
  FIBO(06) =     8 (    8 ticks) (    4 ms)
  FIBO(07) =    13 (    9 ticks) (    5 ms)
  FIBO(08) =    21 (   11 ticks) (    6 ms)
  FIBO(09) =    34 (   14 ticks) (    8 ms)
  FIBO(10) =    55 (   19 ticks) (   11 ms)
  FIBO(11) =    89 (   26 ticks) (   14 ms)
  FIBO(12) =   144 (   38 ticks) (   21 ms)
  FIBO(13) =   233 (   58 ticks) (   32 ms)
  FIBO(14) =   377 (   90 ticks) (   50 ms)
  FIBO(15) =   610 (  143 ticks) (   79 ms)
  FIBO(16) =   987 (  226 ticks) (  126 ms)
  FIBO(17) =  1597 (  363 ticks) (  202 ms)
  FIBO(18) =  2584 (  584 ticks) (  324 ms)
  FIBO(19) =  4181 (  941 ticks) (  523 ms)
  FIBO(20) =  6765 ( 1519 ticks) (  844 ms)
  FIBO(21) = 10946 ( 2455 ticks) ( 1364 ms)
  FIBO(22) = 17711 ( 3969 ticks) ( 2205 ms)
  FIBO(23) = 28657 ( 6418 ticks) ( 3565 ms)
  FIBO(24) = 46368 (10380 ticks) ( 5766 ms)
  '------------------------------------------------

}}

con
  _clkmode = xtal1 + pll16x 'hw#_clkmode
  _xinfreq = 80_000_000 'hw#_xinfreq

#define JAVAEE
#define CACHE

' XRUNNER does not work with SingleSerial
'#define XRUNNER
#ifdef  XRUNNER
#define FDX
#endif

' BMAD enable PASM debugger
'#define BMAD
#ifdef  BMAD
#ifndef FDX
#define FDX
#endif
#endif

' Use FDX for development. SingleSerial is too slow.
#ifndef FDX
#define FDX
#endif

' use to enable HIGH EE Fetch address $8000
#define HIEE

'#define OPCODES

obj

#ifdef FDX
  sx : "FullDuplexSingleton"
#else
  sx : "SingleSerial"
#endif

  pj : "pjvmInterpreterCOG"
  pjd: "pjdata"
  pcn: "pjcon"
  pn : "pnative"
#ifdef GCOLLECT
  pm : "pjmemory"               ' best fit list power of 2 size
#endif
#ifdef JAVAEE
  ee : "PasmA0EEPROM"             ' will use for reading java code from eeprom
#endif

'  hw : "properties"             ' contains user preferences

pub main | progptr, ctrlp, cmdptr, vps
{{
Startup the JVM interpreter
}}
#ifdef JAVAEE
  ee.start(28,29)               ' start EEPROM driver
#endif

#ifdef FDX
  sx.start(31,30,0,115200)
#else
  sx.init(31,30,19200)
#endif
  waitcnt(clkfreq+cnt)

{
#ifdef FDX
  sx.start(hw#UCRXP,hw#UCTXP,0,hw#UCBAUD)
#else
  sx.init(hw#UCRXP,hw#UCTXP,0,hw#UCBAUD)
#endif
  waitcnt(clkfreq*hw#UCWAIT+cnt)
}
  'sx.str(string($d,"Hello JVM",$d))

  ctrlp := pjd.getControlPtr
  cmdptr := pjd.getCmdPtr

#ifndef JAVAEE
  progptr := pj.getProgPtr
  longmove(pj.getProgBuffer, progptr, pj.getProgSize/4)
#endif
  progptr := pj.getProgBuffer

  pjd.init(progptr,pj.getJavaCodeSize,1,false)
#ifdef GCOLLECT
  pm.init(pjd.getHS-2)
#endif

  vps := 3
#ifdef XRUNNER
  cognew(run(progptr, cmdptr, ctrlp), @xstack)
  vps--
#endif
#ifdef BMAD
  cognew(handler, @hstack)     ' for BMA debug only
  vps--
#endif

  pn.start(vps)                 ' start virtual peripherals
  pj.start(cmdptr)              ' start jvm engine

#ifndef BMAD
  handler
#else
  pj.debug
#endif

var long hstack[$80]

pub handler | cmd, tmp, cmdptr, ctlptr
{{
handler loop for JVM services
}}
  cmdptr := pjd.getCmdPtr
  ctlptr := pjd.getControlPtr

  repeat

    handlemsg(ctlptr, cmdptr)

pri handlemsg(ctlptr, cmdptr)
{{
handle memory allocations, natives, and exceptions
}}

#ifdef GCOLLECT
    if (long[ctlptr] > 0) and (long[ctlptr] & pj#MALLOC_BIT)
      tmp := long[ctlptr] & $3fff
      pm.alloc(tmp)
      long[ctlptr] := -1
#endif

    if (long[cmdptr] & pj#EXCPTN_BIT)

      sx.str(string($d,"Exception "))
      sx.hex(long[cmdptr],4)
      sx.str(string(" at "))
      sx.hex(pjd.getPC,4)
      sx.tx($d)
      printPC(cmdptr,0)
      dumpStack(0)
      dumpStatic
      dumpHeap
      'long[cmdptr] := 0
      repeat ' stall forever

    elseif (long[cmdptr] & $100) == $100

      pn.doNative(long[cmdptr] & $3f)
      pjd.incrPC(3)
      long[cmdptr] := 0
{
    elseif long[cmdptr]
        sx.str(string($d,"Unknown message: "))
        sx.hex(long[cmdptr],8)
        sx.tx($d)
        long[cmdptr] := 0
'}

#ifdef XRUNNER

var long xstack[$80]

pub run(progptr, cmdptr, ctrlp) | mpc, opc, trace, n, mp, op1, op2, brk, tmp
  trace~
  trace~~

  opc := 0
  mpc := pjd.getPC

  long[ctrlp] := 0
'{
  waitcnt(clkfreq+cnt)
  printPC(cmdptr,progptr)
  dumpStack(progptr)
  dumpStatic
  dumpHeap
'}
  repeat

    long[ctrlp] := -1   ' step
'    long[ctrlp] := 0    ' step
    repeat
{
      if long[ctrlp]
        sx.str(string($d,"ControlPtr "))
        sx.hex(long[ctrlp],8)
'}
    while long[ctrlp]

    mpc := pjd.getPC
    if(opc <> mpc)
      printPC(cmdptr,progptr)
      'dumpStack(progptr)
      'dumpStatic
      dumpHeap
'      if (mpc & $7fff) == $c78
'        repeat
#ifdef GCOLLECT
      case byte[pjd.getPC+progptr]
        pcn#Jem_NEW, pcn#Jem_NEWARRAY, pcn#Jem_ANEWARRAY:
          printPC(cmdptr,progptr)
          pm.show
          dumpStack(progptr)
          dumpStatic
          dumpHeap
        pcn#Jem_INVOKESTATIC:
          mp := (byte[pjd.getPC+progptr+1]<<8) | byte[pjd.getPC+progptr+2]
          if byte[mp+progptr+2] & $80
            dumpStack(progptr)
            dumpStatic
            dumpHeap
            sx.str(string($d,"################",$d))
#endif

      opc := mpc

      'handlemsg(ctrlp, cmdptr)

#endif
'XRUNNER

pub dumpStack(progptr) | n, mpc, bp
  sx.str(string($d,"Stack "))
  bp  := pjd.getBP
  mpc := pjd.getSP-bp+2
  repeat n from 0 to 32 step 2
    sx.hex(bp+n,4)
    sx.tx(" ")
  repeat n from 0 to mpc step 2
    if n // 32 == 0
      sx.tx($d)
      sx.tx(" ")
      sx.hex(bp+n,4)
    sx.tx(" ")
    sx.hex(swp16(word[progptr+n+bp]),4)

pub dumpStatic | n, hs,stap
  stap := pjd.getStaticPtr
  hs := pjd.getHB - stap
  repeat n from 0 to hs step 2
    if (n // 32) == 0
      sx.tx($d)
      sx.tx("s")
      sx.hex(stap+n,4)
    sx.tx(" ")
    sx.hex(word[stap+n],4)

pub dumpHeap | n, hs
  hs := pjd.getHS-pjd.getHP
  if hs < 2
     return
{
  sx.tx($d)
  sx.hex(pjd.getHP,4)
  repeat n from 0 to hs step 1
    if ((pjd.getHP+n) // 32) == 0
      sx.tx($d)
      sx.hex(pjd.getHP+n,4)
    if byte[pjd.getHP+n] > 15
      sx.tx(byte[pjd.getHP+n])
    else
      sx.tx(".")
}
  sx.tx($d)
  sx.tx("h")
  sx.hex(pjd.getHP,4)
  repeat n from 0 to hs step 2
    if ((pjd.getHP+n) // 32) == 0
      sx.tx($d)
      sx.tx("h")
      sx.hex(pjd.getHP+n,4)
    sx.tx(" ")
    sx.hex(swp16(word[pjd.getHP+n]),4)
  sx.tx($d)


pri swp16(val)
  result := val & $ff
  result <<=8
  val >>=8
  result |= val

pri printPC(cmdptr,progptr)
  sx.tx($d)
  sx.str(string("PC "))
  sx.hex(pjd.getPC,4)
  sx.tx(" ")
#ifdef OPCODES
  padstr(pcn.OpcodeNamePtr(pjd.readProgramByte(pjd.getPC)), 14)
  sx.tx(" ")
#endif
  sx.hex(pjd.readProgramByte(pjd.getPC),2)
  sx.tx(" ")
  sx.hex(pjd.readProgramByte(pjd.getPC+1),2)
  sx.tx(" ")
  sx.hex(pjd.readProgramByte(pjd.getPC+2),2)
  sx.str(string(" M"))
  sx.hex(pjd.getMP,4)
  sx.str(string(" H"))
  sx.hex(pjd.getHP,4)
  sx.str(string(" b"))
  sx.hex(pjd.getSB,4)
  sx.str(string(" B"))
  sx.hex(pjd.getSB,4)
  sx.str(string(" F"))
  sx.hex(pjd.getFP,4)
  sx.str(string(" S"))
  sx.hex(pjd.getSP,4)
  dumpWipFrame(progptr)

pub padstr(str,len) | n
  len -= strsize(str)
  if len > 0
    repeat n from 0 to len
      sx.tx(" ")
  sx.str(str)

pri dumpWipFrame(progptr) | n, sp
  sp := pjd.getSP
  repeat n from 0 to sp-pjd.getFP-2 step 2
    sx.tx(" ")
    sx.hex(swp16(word[n+pjd.getFP]),4)



